#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "defines.h"
#include "video_high.h"
#include "video_low.h"
#include "game.h"

/* defines */

#define CENTER_X            (ORIG_SCREEN_WIDTH / 2 - 1)

/* external variables */

extern FixedPoint Sintable[ANGLES+ANGLES/4];
extern S16 Pixelangle[ORIG_SCREEN_WIDTH];
extern Finetangent[FINEANGLES/4];
extern PL_POINT_TYPE Player;
extern U8 Map[MAP_HEIGHT][MAP_WIDTH];
extern STATIC_OBJECT_TYPE    StaticObjectList[MAXSTATS];
extern STATIC_OBJECT_TYPE    *LastStaticObj;

/* global variables */

U8 PointVisable[MAP_WIDTH][MAP_HEIGHT];
VISABLE_OBJECT_TYPE    VisableList[MAX_VISABLE];
S16 WallHeightScaled[SCREEN_WIDTH];

/* local variables */

static S32          xintercept,yintercept;
static FixedPoint   viewsin,viewcos;
static POINT_TYPE   gridstep;
static FixedPoint   *Costable = Sintable+(ANGLES/4);
static FixedPoint   viewx;
static FixedPoint   viewy;
static S16          viewangle;

/* local prototypes */

static void CalcHeight(U16 *SliceHeight);
static void DrawSlice(S16 x, COLOUR_TYPE Color, U16 SliceHeight);
static void HitHorizWall(S16 pixx, COLOUR_TYPE Color);
static void HitVertWall(S16 pixx, COLOUR_TYPE Color);
static void CalcViewVariables(void);
static COLOUR_TYPE SelectWallColour(U16 Wall);
static void Draw3D(void);
static void DrawScaledSprites(void);
static U8 TransformTile(S32 TileX, S32 TileY, S16 *DispX, S16 *DispHeight);

/*
===============================================================================
 \brief Draw 3d world and places sprites onto it
 \param None
 \return None
===============================================================================
*/
void ThreeDRefresh(void)
{
    /* clear all player visable points  */
    memset(PointVisable,0,(MAP_WIDTH * MAP_HEIGHT)); 
    /* update background for this frame */
    DrawBackGround();
    /* convert from players coords to vector coords */ 
    CalcViewVariables();
    /* draw 3D level and update visable points player can see */
    Draw3D();
    /* Draw player visable sprites into screen */     
    DrawScaledSprites();
    
    Redraw();
}

/*
===============================================================================
 \brief Work out the display height and position for sprite passed 
 \param TileX: X location of object on map
 \param TileY: Y location of object on map
 \param DispX: holds calculated screen X position for sprite
 \param DispHeight: holds calculated screen height for sprite
 \return if object is close enough to be grabbed return one
===============================================================================
*/
static U8 TransformTile(S32 TileX, S32 TileY, S16 *DispX, S16 *DispHeight)
{
    FixedPoint gx,gy,gxt,gyt,nx,ny;   
       
    /* calculate the magnitude of the distances for x and y */
    gx = ((S32)TileX<<MAP_TO_GRID_SHIFT)+ 0x8000 - viewx;
    gy = ((S32)TileY<<MAP_TO_GRID_SHIFT)+ 0x8000 - viewy;
    
    /* convert the xmag and ymag values into equivalent x and y of the polar */
    /* coordinates                                                           */
    gxt = FixedMul(gx,viewcos);
    gyt = FixedMul(gy,viewsin);
     /* add x and y together to get polar distance, note y increases  */ 
    /* downward so x - y actually adds x + y as x - (-y) is +         */
    nx = gxt-gyt-0x2000;                  /* 0x2000 is size of object */

    /* calculate newy */
    gxt = FixedMul(gx,viewsin);
    gyt = FixedMul(gy,viewcos);
    ny = gyt+gxt;

    /* calculate height / perspective ratio */
    if(nx < MINDIST)                 
    {
        *DispHeight = 0;    /* too close, don't overflow the divide */
    }
    else
    {
        *DispX = (S16)(CENTER_X + ny * PROJECTION_RATIO / nx);
        *DispHeight = (S16)(PROJECTION_RATIO_SCALED / nx);
    }
    
    /* see if the object is close enough to be grabbed */
    if((nx < GRID_SIZE) && (ny > -GRID_SIZE/2) && (ny < GRID_SIZE/2))
    {
        return 1;
    }
    else
    {
        return 0;
    }          
}

/*
===============================================================================
 \brief checks to see if sprites needs to be drawn to screen 
 \param None
 \return None
===============================================================================
*/
static void DrawScaledSprites(void)
{
    STATIC_OBJECT_TYPE *StaticObjectPtr;
    VISABLE_OBJECT_TYPE *VisableListPtr = VisableList;
    S32 NumVisable;    
       
    for(StaticObjectPtr = StaticObjectList; StaticObjectPtr != LastStaticObj ; StaticObjectPtr++)
    {
                           
        if(StaticObjectPtr->ShapeNum == -1)
        {
            continue;    /* SCP has already been picked up */
        }
        
        if(*(StaticObjectPtr->VisableSpot) == 0)
        {
            continue;    /* SCP not visable */
        }
        
        /* find static objects distance from player */
        if(TransformTile(StaticObjectPtr->TileX,StaticObjectPtr->TileY
                                ,&VisableListPtr->ViewX,&VisableListPtr->ViewHeight) == 1)
        {
            /* object is very close to player so grab it! */
            GetSCP();
            
            /* object has been taken */
            StaticObjectPtr->ShapeNum = -1;
            continue;    
        }
        
        if(VisableListPtr->ViewHeight == 0)
        {
            continue;    /* to close to the object */
        }
        
        /* add static shape number to visable list */
        VisableListPtr->ShapeNum = StaticObjectPtr->ShapeNum;
        VisableListPtr++;
              
    }
    
    NumVisable = (S32) (VisableListPtr-&VisableList[0]);
    
    if(NumVisable == 0)
    {
        return;    /* no visable objects */
    }
    
    ScaleSprites(NumVisable, SCALING);       
}

/* 
================================================================
=
= Function: Draw3D
=
= Description:
=
= Raycast simple 3d level onto screen
=
================================================================
*/
static void Draw3D(void)
{
        U32         xpartialleft;
        U32         xpartialright;
        U32         ypartialup;
        U32         ypartialdown;
        S32         xstep,ystep;
        U16         xspot,yspot;
        U32         xpartial,ypartial;
        U16         xgrid,ygrid;
        U16         gridhit;
        S16         midangle;
        COLOUR_TYPE Color;
        S16         pixx;
        
        midangle = viewangle*(FINEANGLES/ANGLES);

        /* distance to left side of grid */
        xpartialleft = viewx & (BIN16-1);
        /* distance to right side of grid */
        xpartialright = BIN16 - xpartialleft;
        /* distance to top of grid */
        ypartialup = viewy & (BIN16-1);
        /* distance to bottom of grid */
        ypartialdown = BIN16 - ypartialup;

        /* calculate each ray within the scene */
        for(pixx = 0 ; pixx < ORIG_SCREEN_WIDTH ; pixx++)
        {
            S16 angl = midangle + Pixelangle[pixx];

            if(angl<0)
            {
                angl+=FINEANGLES;
            }
            if(angl>=3600)
            {
                angl-=FINEANGLES;
            }

            if(angl<900)
            {
                /* next grid location from present location */
                gridstep.x = 1;
                gridstep.y = -1;
                /* x or y distance we are going to add for this grid step */
                xstep = Finetangent[900-1-angl];
                ystep = -Finetangent[angl];
                /* distance to grid sides for this angle */
                xpartial = xpartialright;
                ypartial = ypartialup;
            }
            else if(angl<1800)
            {
                gridstep.x = -1;
                gridstep.y = -1;
                xstep = -Finetangent[angl-900];
                ystep = -Finetangent[1800-1-angl];
                xpartial = xpartialleft;
                ypartial = ypartialup;
            }
            else if(angl<2700)
            {
                gridstep.x = -1;
                gridstep.y = 1;
                xstep = -Finetangent[2700-1-angl];
                ystep = Finetangent[angl-1800];
                xpartial = xpartialleft;
                ypartial = ypartialdown;
            }
            else if(angl<3600)
            {
                gridstep.x = 1;
                gridstep.y = 1;
                xstep = Finetangent[angl-2700];
                ystep = Finetangent[3600-1-angl];
                xpartial = xpartialright;
                ypartial = ypartialdown;
            }

            /* work out first y intercept location of the vertial ray */
            yintercept = FixedMul(ystep,xpartial) + viewy;
            /* next x grid location vertical ray will be in */
            xgrid = (viewx >> MAP_TO_GRID_SHIFT) + gridstep.x;
            /* next grid location vertical ray will be in */
            xspot = (U16)(xgrid + (((U32)yintercept >> MAP_TO_GRID_SHIFT) << MAP_WIDTH_SHIFT));

            /* work out first x intercept location of the horizontal ray */
            xintercept = FixedMul(xstep,ypartial) + viewx;
            /* next y grid location horizontal ray will be in */
            ygrid = (viewy >> MAP_TO_GRID_SHIFT) + gridstep.y;
            /* next grid location horizontal ray will be in */
            yspot = (U16)(((U32)xintercept >> MAP_TO_GRID_SHIFT) + (ygrid << MAP_WIDTH_SHIFT));

            /* vertical ray casting */
            do
            {
                if(gridstep.y == -1 && (yintercept >> 16) <= ygrid) goto horizentry;
                if(gridstep.y == 1 && (yintercept >> 16) >= ygrid) goto horizentry;
vertentry:
                /* get value of grid that ray has been cast at */
                gridhit = ((U8 *)Map)[xspot];

                /* does location of ray hold any wall value? */
                if(gridhit != 0)
                {
                    /* choose wall colour */
                    Color = SelectWallColour(gridhit);
                    
                    xintercept = xgrid << MAP_TO_GRID_SHIFT;
                    ygrid = (S16) (yintercept >> MAP_TO_GRID_SHIFT);
                    HitVertWall(pixx, Color);
                    break;
                }

                /* mark this point as visable to player */
                *((U8 *)PointVisable + xspot) = 1;
                /* and move ray to next grid coord's */
                xgrid += gridstep.x;
                yintercept += ystep;
                xspot = (U16)(xgrid + (((U32)yintercept >> 16) << MAP_WIDTH_SHIFT));
            }while(1);

            /* cast next ray */
            continue;

            /* horizontal ray casting */
            do
            {
                if(gridstep.x == -1 && (xintercept >> 16) <= xgrid) goto vertentry;
                if(gridstep.x == 1 && (xintercept >> 16) >= xgrid) goto vertentry;
horizentry:
                /* get value of grid that ray has been cast at */
                gridhit=((U8 *)Map)[yspot];

                /* does location of ray hold any wall value? */
                if(gridhit != 0)
                {
                    /* choose wall colour */
                    Color = SelectWallColour(gridhit + 8);
                    
                    yintercept = ygrid << MAP_TO_GRID_SHIFT;
                    xgrid = (S16) (xintercept >> MAP_TO_GRID_SHIFT);
                    HitHorizWall(pixx, Color);
                    break;
                }
                
                /* mark this point as visable to player */
                *((U8 *)PointVisable + yspot) = 1;
                /* and move ray to next grid coord's */
                ygrid += gridstep.y;
                xintercept += xstep;
                yspot = (U16)(((U32)xintercept >> MAP_TO_GRID_SHIFT) + (ygrid << MAP_WIDTH_SHIFT));
            }while(1);

        }/* end of for loop */
}

/* 
================================================================
=
= Function: HitHorizWall
=
= Description:
=
= Ray hit wall on Horizontal line of grid so calculate distance 
= and display wall height.
=
================================================================
*/
static void HitHorizWall(S16 pixx, COLOUR_TYPE Color)
{
    U16 SliceHeight;

    if(gridstep.y == -1)
    {
        yintercept += GRID_SIZE;
    }

    CalcHeight(&SliceHeight);
    DrawSlice(pixx, Color, SliceHeight);
}

/* 
================================================================
=
= Function: HitVertWall
=
= Description:
=
= Ray hit wall on vertical line of grid so calculate distance 
= and display wall height.
=
================================================================
*/
static void HitVertWall(S16 pixx, COLOUR_TYPE Color)
{
    U16 SliceHeight;

    if(gridstep.x == -1)
    {
        xintercept += GRID_SIZE;
    }

    CalcHeight(&SliceHeight);
    DrawSlice(pixx, Color, SliceHeight);
}

/* 
================================================================
=
= Function: CalcHeight
=
= Description:
=
= Calculates total distance that the ray traveled in x / y 
= directions then convert these distances into the equivalent 
= x and y polar values. Then it the adds the two values to
= get total vector distance travelled by ray. It then uses
= this distance (z) to work out the correct wall height to 
= display. 
=
================================================================
*/
static void CalcHeight(U16 *SliceHeight)
{
    S32 XMag;
    S32 YMag;
    FixedPoint XofPolar;
    FixedPoint YofPolar;
    FixedPoint Z_Polar;
    
    /* calculate the magnitude of the distances for x and y */
    XMag = xintercept - viewx;
    YMag = yintercept - viewy;
    
    /* convert the xmag and ymag values into equivalent x and y of the polar */
    /* coordinates                                                           */
    XofPolar = FixedMul(xintercept - viewx, viewcos);
    YofPolar = FixedMul(yintercept - viewy, viewsin);
    
    /* add x and y together to get polar distance, note y increases  */ 
    /* downward so x - y actually adds x + y as x - (-y)             */
    Z_Polar = XofPolar - YofPolar;

    *SliceHeight = PROJECTION_RATIO_SCALED / Z_Polar;
}

/* 
================================================================
=
= Function: DrawSlice
=
= Description:
=
= Draws the wall height onto the screen centred around the
= screens midpoint.
=
================================================================
*/
static void DrawSlice(S16 X, COLOUR_TYPE Colour, U16 SliceHeight)
{
    S32 DrawStart;
    S32 i;

    /* clip height to screen height if needed */
    if(SliceHeight > (ORIG_SCREEN_HEIGHT - STATUSLINES))
    {
        SliceHeight = (ORIG_SCREEN_HEIGHT - STATUSLINES);
    }
    
    /* store height info for comparing against sprite sizes later on */
    for(i = 0 ; i < SCALING ; i++)
    {
        WallHeightScaled[(X * SCALING) + i] = SliceHeight * SCALING;
    }

    DrawStart = (ORIG_SCREEN_CENTRE - (STATUSLINES / 2)) - (SliceHeight / 2);  
  
    /* draw the pixels of the stripe as a vertical line */
    DrawVertLine(X, DrawStart, SliceHeight, Colour, SCALING);
}

/* 
================================================================
=
= Function: CalcViewVariables
=
= Description:
=
= setup view variables before updating frame
=
================================================================
*/
static void CalcViewVariables(void)
{    
    viewangle = Player.Angle;
    viewsin = Sintable[viewangle];
    viewcos = Costable[viewangle]; 
     
    /* convert players position into screen x ,y coords */
    viewx = Player.X - FixedMul(FOCAL_LENGTH,viewcos);
    viewy = Player.Y + FixedMul(FOCAL_LENGTH,viewsin);
}

/* 
================================================================
=
= Function: SelectWallColour
=
= Description:
=
= Chooses the correct wall colour for the curent wall hit by the
= ray
=
================================================================
*/
static COLOUR_TYPE SelectWallColour(U16 Wall)
{
    COLOUR_TYPE Colour;         
            
    switch(Wall)
    {
        case 1:     /* Navy */
            Colour.r = 0;
            Colour.g = 0;
            Colour.b = 128;
        break;

        case 2:     /* green */
            Colour.r = 0;
            Colour.g = 128;
            Colour.b = 0;
        break;

        case 3:     /* Teal */
            Colour.r = 0;
            Colour.g = 128;
            Colour.b = 128;
        break;
        
        case 4:    /* Maroon */
            Colour.r = 128;
            Colour.g = 0;
            Colour.b = 0;
        break;
        
        case 5:    /* Purple */
            Colour.r = 128;
            Colour.g = 0;
            Colour.b = 128;
        break;
        
        case 6:    /* Olive */
            Colour.r = 128;
            Colour.g = 128;
            Colour.b = 0;
        break;
        
        case 7:    /* Gray */
            Colour.r = 128;
            Colour.g = 128;
            Colour.b = 128;
        break;
        
        case 8:    /* Silver */
            Colour.r = 192;
            Colour.g = 192;
            Colour.b = 192;
        break;
        
        case 9:    /* Blue */
            Colour.r = 0;
            Colour.g = 0;
            Colour.b = 255;
        break;
        
        case 10:    /* Lime */
            Colour.r = 0;
            Colour.g = 255;
            Colour.b = 0;
        break;
        
        case 11:    /* Aque */
            Colour.r = 0;
            Colour.g = 255;
            Colour.b = 255;
        break;
        
        case 12:    /* Red */
            Colour.r = 255;
            Colour.g = 0;
            Colour.b = 0;
        break;
        
        case 13:    /* Pink */
            Colour.r = 255;
            Colour.g = 0;
            Colour.b = 255;
        break;
        
        case 14:    /* Yellow */
            Colour.r = 255;
            Colour.g = 255;
            Colour.b = 0;
        break;
        
        case 15:    /* White */
            Colour.r = 255;
            Colour.g = 255;
            Colour.b = 255;
        break;    
    }
    
    return Colour;                
};
