#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "defines.h"
#include "inputs.h"
#include "video_low.h"
#include "video_high.h"
#include "player.h"
#include "time.h"
#include "act1.h"
#include "draw.h"

/* defines */

#define EXTRA_ZERO     0
#define GAME_OVER      "GAME OVER"
#define COMPLETE       "LEVEL COMPLETE" 

/* external variables */

extern PL_POINT_TYPE Player;

/* global variables */

GAMESTATE_TYPE GameState;

/* local variables */

/* None */

/* local prototypes */

static void UpdateStatusBar(void);
static void DrawLevel(void);
static void DrawNumber(U16 X, U16 Y, S32 Number, U8 Scaling);
static void DrawSCP(void);
static void DrawLives(void);
static void DrawHealth(void);
static void DrawTime(void);
static void InitGameState(void);

/*
===============================================================================
 \brief Draws game world and updates players position
 \param None
 \return None
===============================================================================
*/
U8 GameLoop(void)
{
    INPUT_TYPE UserControls;
    memset(&UserControls,0,sizeof(UserControls));
    
    /* Setup inital game state */
    InitGameState();
    /* Load static objects */
    InitStaticObjects();
                
    /* Draw first interim scene */
    DrawPic(0, 0, InterRimPic1, SCALING);
    /* Draw status bar and fill with correct values */
    UpdateStatusBar();
    /* Give user time to read the display then start game */
    UserControls.Exit = DelaySecondsWithExit(20);
    
    /* Reset game time before we enter loop */
    SetGameTime();    

    /* display simple 3d level until user decides to exit */
    while((UserControls.Exit == 0) && (GameState.PlayState == Playing))
    {
        /* clear last user input */
        memset(&UserControls,0,sizeof(UserControls));
        /* then see if user was pressed a key */
        GetUserInput(&UserControls);
        /* move player if user pressed the arrow keys */
        MovePlayer(&UserControls);
        /* do raycasting of level */
        ThreeDRefresh();
        /* Update time left to get all SCP */
        UpdateGameTime();
        DrawTime();
        /* update health and see if player has died */
        DrawHealth();
        /* wait so it consumes less processing time */
        DelayMs(5); 
    }
    
    if(GameState.PlayState == Died)
    {
        Message(GAME_OVER);
        Redraw();
        return DelaySecondsWithExit(5);
    }
    
    if(GameState.PlayState == Victory)
    {
        Message(COMPLETE);
        Redraw();
        return DelaySecondsWithExit(5);
    }
    
    return UserControls.Exit;    
}

/*
===============================================================================
 \brief check to see if we have picked up all the SCP and have completed game
 \param None
 \return None
===============================================================================
*/
void GetSCP(void)
{
    GameState.Scp++;
    DrawSCP();
    
    if(GameState.Scp == 5)
    {
        GameState.PlayState = Victory;                 
    }
}

/*
===============================================================================
 \brief Setup inital game state before entry into game
 \param None
 \return None
===============================================================================
*/
static void InitGameState(void)
{
    GameState.LevelNum = Level1;
    GameState.Health = 100;
    GameState.Lives = 1;
    GameState.Scp = 0;
    GameState.PlayState = Playing;
    switch(GameState.SkillLevel)
    {
        case Easy:
            GameState.TimeCount = 150;    
        break;
        
        case Medium:
            GameState.TimeCount = 120;    
        break;  
        
        case Hard:
            GameState.TimeCount = 90;    
        break;  
        
        case Crazy:
            GameState.TimeCount = 60;    
        break;                              
    };
    
    /* define players start position x = 8.5 y = 8.5 angle = 0 */
    Player.X = 8.5 * GRID_SIZE;
    Player.Y = 8.5 * GRID_SIZE;
    Player.Angle = 0;     
}

/*
===============================================================================
 \brief Draw status bar and fill with correct value
 \param None
 \return None
===============================================================================
*/
static void UpdateStatusBar(void)
{
    DrawPic(0,(ORIG_SCREEN_HEIGHT - STATUSLINES),StatusBarPic, SCALING);
    DrawLevel();
    DrawSCP();
    DrawLives();
    DrawHealth();
    DrawTime();
    Redraw();
}

/*
===============================================================================
 \brief Draw time remaining before room becomes toxic
 \param None
 \return None
===============================================================================
*/
static void DrawTime(void)
{
    U8 Minutes = GameState.TimeCount / 60;       
    U8 Seconds = GameState.TimeCount % 60;     
    DrawNumber(275, 16, Minutes, SCALING);
    DrawPic(283 ,ORIG_SCREEN_HEIGHT - (STATUSLINES -  16),ColonPic,SCALING);
    if(Seconds < 10)
    {
        /* Add leading zero */
        DrawNumber(288, 16, EXTRA_ZERO, SCALING);
        DrawNumber(296, 16, Seconds, SCALING);
    }
    else
    {
        DrawNumber(288, 16, Seconds, SCALING);    
    }       
}

/*
===============================================================================
 \brief Draw players health into status bar
 \param None
 \return None
===============================================================================
*/
static void DrawHealth(void)
{
    /* do we have less then a 1.5 minutes to go ?*/   
    if(GameState.TimeCount < 100)
    {
        GameState.Health = GameState.TimeCount;
        /* Add leading zero */
        DrawNumber(210, 16, EXTRA_ZERO, SCALING);
        if(GameState.Health < 10)
        {
            DrawNumber(218, 16, EXTRA_ZERO, SCALING);
            DrawNumber(226, 16, GameState.Health, SCALING);
        }
        else
        {
            DrawNumber(218, 16, GameState.Health, SCALING);
        }
        if(GameState.Health == 0)
        {
            GameState.PlayState = Died;
        }
    }
    else
    {   
        DrawNumber(210, 16, GameState.Health, SCALING);
    }       
}

/*
===============================================================================
 \brief Draw Number of players lives lift into status bar
 \param None
 \return None
===============================================================================
*/
static void DrawLives(void)
{
    DrawNumber(145, 16, GameState.Lives, SCALING);       
}

/*
===============================================================================
 \brief Draw Number of science crop personal rescued into status bar
 \param None
 \return None
===============================================================================
*/
static void DrawSCP(void)
{
    DrawNumber(90, 16, GameState.Scp, SCALING);       
}

/*
===============================================================================
 \brief Draw Level info into status bar
 \param None
 \return None
===============================================================================
*/
static void DrawLevel(void)
{
    DrawNumber(25, 16, GameState.LevelNum + 1, SCALING);       
}

/*
===============================================================================
 \brief Print Number passed into the status bar at x, y
 \param X: start coordinate of number
 \param Y: start coordinate of number
 \param Number: Number to print to screen
 \param Scaling: scaling factor (1 if not required) 
 \return None
===============================================================================
*/
static void DrawNumber(U16 X, U16 Y, S32 Number, U8 Scaling)
{
    U32 Length;
    U32 i;
    char Str[20];

    /* convert number to string */
    sprintf(Str,"%d",Number);

    Length = (U32) strlen(Str);

    /* draw number */
    for(i = 0; i < Length; i++)
    {
        DrawPic(X ,ORIG_SCREEN_HEIGHT - (STATUSLINES -  Y),((Str[i]-'0') + NumZeroPic),Scaling);
        X += CHAR_WIDTH;
    }
}
