#include <stdio.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include "types.h"
#include "defines.h"
#include "video_low.h"

/* defines */

#define BIT_MASK 0x01
#define TITLE    "Dark Nights"
#define BLACK    0
#define GREY     0x808080

/* external variables */

extern const U32 Font[192];
extern const PIC_TYPE *Pics[NumberPics];
extern VISABLE_OBJECT_TYPE    VisableList[MAX_VISABLE];
extern S16 WallHeightScaled[SCREEN_WIDTH];

/* global variables */

U16 CursorLocX = 0;
U16 CursorLocY = 0;

/* local variables */

static SDL_Surface *scr;
static COLOUR_TYPE TextColour = {0,0,0};
static COLOUR_TYPE BackColour = {255,255,255};
static SDL_Rect UpperArea = {0,0,SCREEN_WIDTH, (SCREEN_HEIGHT - (STATUSLINES * SCALING)) / 2};
static SDL_Rect LowerArea = {0, (SCREEN_HEIGHT - (STATUSLINES * SCALING)) / 2, SCREEN_WIDTH, (SCREEN_HEIGHT - (STATUSLINES * SCALING)) / 2};

/* local prototypes */

static void SetRetangle(SDL_Rect *Area, U32 Colour);

/*
===============================================================================
 \brief Draws a pic to screen at x ,y
 \param X: start coordinate of picture  
 \param Y: start coordinate of picture 
 \param PicNum: Picture to draw to screen
 \param Scaling: scaling factor (1 if not required)
 \return None
===============================================================================
*/
void DrawPic(U16 X, U16 Y, GRAPHICS_TYPE PicNum, U8 Scaling)
{    
    U8 *pSource = Pics[PicNum]->PixelData;
    U32 Width = Pics[PicNum]->Width;
    U32 Height = Pics[PicNum]->Height;
    U32 BytesPerPixel = Pics[PicNum]->BytesPerPixel;
    U32 i;
    U32 j;
    U32 k;
    COLOUR_TYPE  Colour;
    U32 ColourSDL;
    U32 XScaled = X * Scaling;
    U32 YScaled = Y * Scaling;
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_LockSurface(scr);
    }
    
    U32 *bufp = (U32 *)scr->pixels + YScaled * scr->w + XScaled;
    
    /* loop until all lines of the image are drawn */
    for(i = 0; i < (Height * Scaling) ; i++)
    {
        if((i % Scaling == 0) && (i != 0))
        {
            /* move source pointer forward to start of next line */  
            pSource += Width * BytesPerPixel;
        }
        
        /* draw one line at a time */
        for(j = 0 ; j <  Width ; j++)
        {  
            Colour.r = (U8)pSource[j * BytesPerPixel];
            Colour.g = (U8)pSource[j * BytesPerPixel + 1];
            Colour.b = (U8)pSource[j * BytesPerPixel + 2];
  
            ColourSDL = SDL_MapRGB(scr->format, Colour.r, Colour.g, Colour.b);
  
            /* Scale image width if needed */
            for(k = 0; k < Scaling ; k++)
            {
                *bufp = ColourSDL;
                bufp++;
            }
        }
        
        /* if source width is not the same width as screen move screen pointer */
        if((Width * Scaling) < scr->w)
        {
            bufp += (scr->w - (Width * Scaling));                
        }
    }
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_UnlockSurface(scr);
    } 
                        
} 

/*
===============================================================================
 \brief Writes a string of chars to screen in TextColour, note does
        not handle control chars at the moment. 
 \param string: message to print to the screen
 \param X: start coordinate of string  
 \param Y: start coordinate of string  
 \param Scaling factor to use for offset coordinates
 \return None
===============================================================================
*/
void MyPrint(const char *String, U16 Xoffset, U16 Yoffset, U8 Scaling)
{
    U8 FontOffset;
    U8 *Byte = NULL;
    U32 i;
    U32 j;
    U32 ColourSDL;
    U32 XoffsetScaled = Xoffset * Scaling;
    U32 YoffsetScaled = Yoffset * Scaling;
    CursorLocX = Xoffset;
    CursorLocY = Yoffset;
        
    /* init pointer to start of screen area */
    U32 *pScreenBuf = (U32 *)scr->pixels;
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_LockSurface(scr);
    }
    
    /* print whole string to screen */   
    while(*String != '\0')
    {
        FontOffset = (*String - 32) * 2;
        String++;
        Byte = (U8 *)&Font[FontOffset];
        
        /* loop until we have printed one whole char */
        for(i = 0 ; i < CHAR_HEIGHT; i++)
        {    
             /* print one line of a char */         
             for(j = 0 ; j < CHAR_WIDTH; j++)
             {
                 if(Byte[i] & (BIT_MASK << j))
                 {
                     ColourSDL = SDL_MapRGB(scr->format, TextColour.r, TextColour.g, TextColour.b);
                     pScreenBuf[((scr->pitch / 4) * (i + YoffsetScaled)) + j + XoffsetScaled] = ColourSDL;   
                 }
                 else
                 { 
                     ColourSDL = SDL_MapRGB(scr->format, BackColour.r, BackColour.g, BackColour.b);
                     pScreenBuf[((scr->pitch / 4) * (i + YoffsetScaled)) + j + XoffsetScaled] = ColourSDL;     
                 }      
             }
        }
        
        /* move to next charactor space */
        pScreenBuf += CHAR_WIDTH;
        CursorLocX += CHAR_WIDTH / Scaling; 
    }
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_UnlockSurface(scr);
    }    
}

/*
===============================================================================
 \brief Draws a single verticle line on screen of the colour passed
 \param X: start point of line in the x plane
 \param Y: start point of line in the y plane
 \param Height: Height of line to draw
 \param Colour: Colour of line to be drawn
 \param Scaling: scaling value (1 if no scaling required )
 \return None
===============================================================================
*/
void DrawVertLine(U32 x, S32 y, U32 Height, COLOUR_TYPE Colour, U8 Scaling)
{
    U32 ScreenWidth = 320 * Scaling;
    U32 y_scaled = y * Scaling;
    U32 x_scaled = x * Scaling;
    U32 HeightScaled = Height * Scaling;
    U32 i;
    U32 ColourSDL; 
     
    U32 *dest = (U32 *)scr->pixels + y_scaled * ScreenWidth + x_scaled;
    
    while(HeightScaled-- != 0)
    {    
        ColourSDL = SDL_MapRGB(scr->format, Colour.r, Colour.g, Colour.b);
        
        /* Scale width if needed */
        for(i = 0; i < Scaling ; i++)
        {
            *dest = ColourSDL;
            dest++;
        }
        
        dest += ScreenWidth - Scaling;
    }
}

/*
===============================================================================
 \brief Draws a single horizontal line on screen of the colour passed
 \param X: start point of line in the x plane
 \param Y: start point of line in the y plane
 \param Width: width of the line
 \param Colour: Colour of line to be drawn
 \param Scaling: scaling value (1 if no scaling required )
 \return None
===============================================================================
*/
void DrawHorzLine(U32 x, S32 y, U32 Width, COLOUR_TYPE Colour, U8 Scaling)
{
    U32 ScreenWidth = 320 * Scaling;
    U32 y_scaled = y * Scaling;
    U32 x_scaled = x * Scaling;
    U32 WidthScaled = Width * Scaling;
    U32 i;
    U32 ColourSDL; 
     
    U32 *dest = (U32 *)scr->pixels + y_scaled * ScreenWidth + x_scaled;
    
    for(i = 0 ; i <  WidthScaled ; i++)
    {    
        ColourSDL = SDL_MapRGB(scr->format, Colour.r, Colour.g, Colour.b);
        *dest = ColourSDL;
        dest++;
    }
}

/*
===============================================================================
 \brief Draws a solid rectangle on screen of the colour passed 
 \param X: start point of outline in the x plane
 \param Y: start point of outline in the y plane
 \param Width: width of rectangle
 \param Height: height of rectangle
 \param Colour: colour of rectangle
 \param Scaling: scaling value (1 if no scaling required )
 \return None
===============================================================================
*/
void DrawRec(U32 x, S32 y, U32 Width, U32 Height, COLOUR_TYPE Colour, U8 Scaling)
{
    U32 ScreenWidth = 320 * Scaling;
    U32 y_scaled = y * Scaling;
    U32 x_scaled = x * Scaling;
    U32 WidthScaled = Width * Scaling;
    U32 HeightScaled = Height * Scaling;
    U32 ColourSDL;
    U32 i;    
     
    U32 *dest = (U32 *)scr->pixels + y_scaled * ScreenWidth + x_scaled;

    while(HeightScaled > 0)
    {
        for(i = 0 ; i <  WidthScaled ; i++)
        {    
            ColourSDL = SDL_MapRGB(scr->format, Colour.r, Colour.g, Colour.b);
            *dest = ColourSDL;
            dest++;
        }
            
        dest += ScreenWidth - WidthScaled;
        HeightScaled--;
    }
}

/*
===============================================================================
 \brief Draw a buffered bmp image into screen scaling if needed
 \param Pointer to source of bitmap to draw
 \param Width of bitmap
 \param Height of bitmap
 \param How many bytes per pixel the bitmap uses
 \param Number of padding bytes used per line
 \param Scaling factor used to display the image
 \return None
===============================================================================
*/
void DrawBmp(U8 *pSource, U32 BitMapWidth, U32 BitMapHeight, U8 BytesPerPixel, U8 Padding, U8 Scaling)
{
    COLOUR_TYPE  Colour;
    U32 *bufp = (U32 *)scr->pixels;
    U32 i;
    U32 j;
    U32 k;
    U32 ColourSDL;
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_LockSurface(scr);
    }
      
    /* loop until all lines of the image are drawn */
    for(i = 0; i < (BitMapHeight * Scaling) ; i++)
    {
        if(i % Scaling == 0)
        {
            /* move pointer back to start of next line */  
            pSource -= BitMapWidth * BytesPerPixel + Padding;
        }
        
        /* draw one line at a time */
        for(j = 0 ; j <  BitMapWidth ; j++)
        {  
            Colour.b = (U8)pSource[j * BytesPerPixel];
            Colour.g = (U8)pSource[j * BytesPerPixel + 1];
            Colour.r = (U8)pSource[j * BytesPerPixel + 2];
  
            ColourSDL = SDL_MapRGB(scr->format, Colour.r, Colour.g, Colour.b);
  
            /* Scale image width if needed */
            for(k = 0; k < Scaling ; k++)
            {
                *bufp = ColourSDL;
                bufp++;
            }
        }
    }
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_UnlockSurface(scr);
    } 
}

/*
===============================================================================
 \brief Try and get a handle to the screen buffer
 \param Width: width of the screen window to open
 \param Height: height of the screen window to open
 \return if we can get get hold a screen window return 0 else we fail return 1
===============================================================================
*/
U8 InitSDLScreenBuf(S32 Width, S32 Height)
{
    S32 ColorDepth = 32;
    
     /* use SDL lib to get access to screen buffer */
    if(SDL_Init(SDL_INIT_EVERYTHING) < 0)
    {
        printf("Unable to init SDL: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }
    
    SDL_WM_SetCaption(TITLE, NULL);

    scr = SDL_SetVideoMode(Width, Height, ColorDepth, SDL_HWSURFACE | SDL_HWPALETTE);
    
    if(scr == NULL)
    {
        printf("Unable to set video: %s\n", SDL_GetError());
        SDL_Quit();
        return(1);
    }
    
    SDL_ShowCursor(SDL_DISABLE);

    return(0);
}

/*
===============================================================================
 \brief Updates the screen.  Has to be called to view new pixels,
        but use only after drawing the whole screen because it's slow.
 \param None
 \return None
===============================================================================
*/
void Redraw(void)
{
    if(SDL_MUSTLOCK(scr))
    {
        SDL_LockSurface(scr);
    }

    SDL_UpdateRect(scr, 0, 0, 0, 0);
    
    if(SDL_MUSTLOCK(scr))
    {
        SDL_UnlockSurface(scr);
    } 
}

/*
===============================================================================
 \brief Clears the retangle area to colour passed
 \param None
 \return None
===============================================================================
*/
static void SetRetangle(SDL_Rect *Area, U32 Colour)
{   
  SDL_FillRect(scr, Area, Colour);
}

/*
================================================================
=
= Function: ScaleShape
=
= Description:
=
= draw shapenum to screen but scaled (using Nearest Neighbor 
= Image Scaling) according to view height and view width
=
================================================================
*/
void ScaleSprites(U8 NumSprites, U8 Scaling)
{
    U32 Scale;
    U32 i;
    S16 NewSpriteSize;
    S16 ActX;
    S32 UpperEdge;
    S32 PixelCount;
    S32 LeftXScreenPic;
    S32 RightXScreenPic;
    S32 j;
    S32 YCount;
    S32 UpperYScreenPic;
    S32 LowerYScreenPic;
    U32 ColourSDL;  
    COLOUR_TYPE  Colour;
    VISABLE_OBJECT_TYPE *VisableListPtr;
    U32 *pScreenBuf;
    U8 *PicSourcePtr;
    U32 OrigSpriteWidth;
    U32 OrigSpriteHeight;
    U32 BytesPerPixel;
    SDL_Surface *SpriteSurface;
    
    /* Part of the bitmap that we want to draw */
    SDL_Rect Source;
    /* Part of the screen we want to draw the sprite to */
    SDL_Rect Destination;
    /* Used to fill surface with default colour */
    SDL_Rect FillSurFace;
    
    FillSurFace.x = 0;
    FillSurFace.y = 0;
    FillSurFace.w = SCREEN_WIDTH;
    FillSurFace.h = SCREEN_HEIGHT;
    
    SpriteSurface = SDL_CreateRGBSurface(SDL_SRCALPHA | SDL_HWSURFACE, SCREEN_WIDTH, SCREEN_HEIGHT, 32, 
                                  0, 0, 0, 0);
    
    SDL_SetColorKey(SpriteSurface, SDL_SRCCOLORKEY, SDL_MapRGB(SpriteSurface->format, '\230', 0, '\210') );
    
    /* fill surface with transparent colour */
    SDL_FillRect(SpriteSurface, &FillSurFace, SDL_MapRGB(SpriteSurface->format, '\230', 0, '\210'));  
    
    for(i = 0; i < NumSprites; i++)
    {
        VisableListPtr = &VisableList[i];         
        PicSourcePtr = Pics[VisableListPtr->ShapeNum]->PixelData;
        OrigSpriteWidth = Pics[VisableListPtr->ShapeNum]->Width;
        OrigSpriteHeight = Pics[VisableListPtr->ShapeNum]->Height;
        BytesPerPixel = Pics[VisableListPtr->ShapeNum]->BytesPerPixel;
          
        NewSpriteSize = VisableListPtr->ViewHeight * Scaling;
        ActX =  VisableListPtr->ViewX * Scaling - NewSpriteSize / 2;
        UpperEdge = ((ORIG_SCREEN_HEIGHT - STATUSLINES) * Scaling) / 2 - (NewSpriteSize / 2);
        
        /* draw from left to right  */
        for(i = 0; i < OrigSpriteWidth ; i++)
        {
            /* work out the position of the left screen pix to start drawing from */  
            PixelCount = i * NewSpriteSize;
            LeftXScreenPic = (PixelCount>>6) + ActX;
            
            /* work out the position of the right sceen pix to draw too */
            PixelCount += NewSpriteSize;
            RightXScreenPic = (PixelCount >> 6) + ActX;
            
            /* make sure we have a pixels to draw */
            if((LeftXScreenPic != RightXScreenPic) && (RightXScreenPic > 0))
            {
                /* make sure left pixel start coord is on screen */
                if(LeftXScreenPic < 0)
                {
                    LeftXScreenPic = 0;
                }
                
                /* make sure right screen pixel is on screen */
                if(RightXScreenPic >= SCREEN_WIDTH)
                {
                    RightXScreenPic = SCREEN_WIDTH - 1;
                } 
                
                /* loop using scaled x start and x end pixel coordinates */
                while(LeftXScreenPic < RightXScreenPic)
                {
                    /* only draw sprite if it is bigger then background wall height */
                    if(WallHeightScaled[LeftXScreenPic] <= NewSpriteSize)
                    {                 
                                    
                        j = 0;
                    
                        /* scale orginal image pixel location to location required by */
                        /* new image size (which is based on z distance)              */
                        YCount = j * NewSpriteSize;
                    
                        /* make sure scaling is in steps of 64 (>>6) as the orginal   */
                        /* image is 64 by 64 and add image start offset               */
                        UpperYScreenPic =(YCount >> 6) + UpperEdge;
                    
                        /* point screen buffer at first pixel to draw on */
                        if(UpperYScreenPic < 0)
                        {
                            pScreenBuf = (U32 *)SpriteSurface->pixels + LeftXScreenPic;
                        }
                        else
                        {
                            pScreenBuf = (U32 *)SpriteSurface->pixels + UpperYScreenPic * SCREEN_WIDTH + LeftXScreenPic; 
                        }
                                     
                        /* loop using orginal ystart and yend coordinates */
                        for(; j < OrigSpriteHeight; j++)
                        {       
                            /* scale orginal image pixel location to location required by */
                            /* new image size (which is based on z distance)              */
                            YCount = j * NewSpriteSize;
                            /* make sure scaling is in steps of 64 (>>6) as the orginal   */
                            /* image is 64 by 64 and add image start offset               */
                            UpperYScreenPic =(YCount >> 6) + UpperEdge; 
                            /* calculate the new end y pixel coodinate based on scaling */
                            /* of image                                                 */
                            YCount += NewSpriteSize;
                            LowerYScreenPic =(YCount >> 6) + UpperEdge;
                        
                            Colour.r = PicSourcePtr[(i + j * OrigSpriteHeight) * BytesPerPixel];
                            Colour.g = PicSourcePtr[(i + j * OrigSpriteHeight) * BytesPerPixel + 1];
                            Colour.b = PicSourcePtr[(i + j * OrigSpriteHeight) * BytesPerPixel + 2];
                            Colour.a = PicSourcePtr[(i + j * OrigSpriteHeight) * BytesPerPixel + 3];
                        
                            ColourSDL = SDL_MapRGBA(SpriteSurface->format, Colour.r, Colour.g, Colour.b, Colour.a);
                        
                            /* check to see if we have some pixels to draw */
                            if((UpperYScreenPic != LowerYScreenPic) && (LowerYScreenPic > 0))
                            {
                                /* make sure start pixel is a positive value */
                                if(UpperYScreenPic < 0)
                                {
                                    UpperYScreenPic = 0;
                                }
                                
                                /* make sure end pixel is not off the end of screen */
                                if(LowerYScreenPic > (ORIG_SCREEN_HEIGHT - STATUSLINES) * Scaling)
                                {
                                    LowerYScreenPic = (ORIG_SCREEN_HEIGHT - STATUSLINES) * Scaling;
                                    j = LowerYScreenPic;
                                }
                            
                                /* colour pixel from ystart to yend */
                                while(UpperYScreenPic < LowerYScreenPic)
                                {                 
                                    *pScreenBuf = ColourSDL;
                                    pScreenBuf += SCREEN_WIDTH;
                                    UpperYScreenPic++;
                                }    
                            }                               
                        }
                    }
                  LeftXScreenPic++;                    
                }
            }
        }        
    }      
   
   /* setup the sprite area to copy from */
   Source.y = UpperEdge;
   Source.x = ActX;
   Source.w = NewSpriteSize;
   
   /* setup the sprite area to copy too */
   Destination.y = UpperEdge;
   Destination.x = ActX;
   Destination.w = NewSpriteSize;
   
   /* check to see if sprite is going to be copy over status bar */
   if(Destination.y + NewSpriteSize >= ((ORIG_SCREEN_HEIGHT - STATUSLINES) * Scaling))
   {
       Source.h = Destination.h = ((ORIG_SCREEN_HEIGHT - STATUSLINES) * Scaling) - Destination.y;      
   }
   else
   {
       Source.h = Destination.h = NewSpriteSize;    
   }

   /* copy only non transparent colours into screen buffer */
   SDL_BlitSurface(SpriteSurface, &Source, scr, &Destination);
   
   /* all done free screen surface */
   SDL_FreeSurface(SpriteSurface);    
}

/*
===============================================================================
 \brief Draw Ceiling and floor into background before raycasting starts
 \param None
 \return None
===============================================================================
*/
void DrawBackGround(void)
{
    SetRetangle(&UpperArea, BLACK);
    SetRetangle(&LowerArea, GREY);         
}

/*
===============================================================================
 \brief Close SDL system and free resources
 \param None
 \return None
===============================================================================
*/
void CloseSDL(void)
{
    /* all done free screen surface */
    SDL_FreeSurface(scr);
    /* close SDL resources */ 
    SDL_Quit();
}
